/**
 *
 * lca.c
 *
 * Implementation of the LCA (Least Common Ancestor) data structure.  This
 * is an implementation of the paper by Bender and Farach-Colton.  This
 * algorithm requires linear pre-processing time and can perform a LCA
 * query in constant time.
 *
 * It depends on a reduction to the +/-1RMQ (Range Minimum Query) problem.
 */
#include <stdlib.h>
#include <stdio.h>
#include "lca.h"
#include "rmq.h"

static node_t *new_node( node_t *left, node_t *right, unsigned int value )
{
  node_t *node = (node_t *) malloc( sizeof( node_t ));

  if ( node )
  {
    node->left  = left;
    node->right = right;
    node->value = value;
  }

  return node;
}

static unsigned int euler_tour( node_t *node, unsigned int step, 
                                unsigned int depth, unsigned int *level, 
                                node_t **euler )
{
  /* Base case */
  if ( node == NULL )
    return step;

  /**
   * In an Euler tour each node is visited c+1 times where c is the number
   * of children.
   */
  if ( node->left != NULL )
  {
    level[step  ] = depth;
    euler[step++] = node;
    
    step = euler_tour( node->left, step, depth + 1, level, euler );
  }
  
  level[step  ] = depth;
  euler[step++] = node;
  
  if ( node->right != NULL )
  {
    step = euler_tour( node->right, step, depth + 1, level, euler );
    
    level[step  ] = depth;
    euler[step++] = node;
  }

  return step;
}

/**
 * When initializing the LCA, we have to compute an Euler tour of the tree,
 * the level of each node, and a representative array.
 */
lca_t *lca_init( node_t *root, unsigned int n )
{
  unsigned int i, tour_len = 2 * n - 1;
  unsigned int *L, *R;
  node_t      **E;
  lca_t        *lca;

  L = (unsigned int *) malloc( tour_len * sizeof( unsigned int ) );
  E = (node_t **)      malloc( tour_len * sizeof( node_t * ) );
  R = (unsigned int *) malloc( n * sizeof( unsigned int ) );

  euler_tour( root, 0, 0, L, E );

  /**
   * Find the index of the first occurence of each node in the euler array
   * by walking the list in reverse order and recording the index of each
   * value 
   */
  i = tour_len;
  while ( i-- )
    R[E[i]->value] = i;

  /**
   * Allocate and fill in the LCA structure using the +/-1 RMQ data
   * structure. 
   */
  lca = (lca_t *) malloc( sizeof( lca_t ));

  lca->E   = E;
  lca->R   = R;
  lca->rmq = pm_rmq_init( L, tour_len );

  return lca;
}

node_t *lca_query( lca_t *lca, unsigned int u, unsigned int v )
{
  unsigned int i = uint_min( lca->R[u], lca->R[v] );
  unsigned int j = uint_max( lca->R[u], lca->R[v] );
  
  return lca->E[ pm_rmq_exec( lca->rmq, i, j ) ];
}

/**
 * A cartesian tree is needed in order to map from the general RMQ problem
 * to the LCA problem.
 */
node_t *build_cartesian_tree( unsigned int *a, unsigned int n )
{
  node_t *curr, *prev, *root = NULL;
  unsigned int i, path_len;

  for ( i = 0; i < n; i++ )
  {
    /* Find the position of a[i] on the right-most path */
    curr = root;
    while ( curr != NULL && a[i] > a[curr->value] )
    {
      prev = curr;
      curr = curr->right;
    }

    /**
     * If the found node is equal to the root, the node node it the root,
     * otherwise hang it off the right subtree of the previous node.
     */
    if ( curr == root )
      root = new_node( curr, NULL, i );
    else
      prev->right = new_node( curr, NULL, i );
  }

  return root;
}
