/**
 *
 * Range Minimum Query algorithms
 *
 * This file contains five implementations of a data structure for
 * solving the Range Minimum Query problem.  Using the notation from Bender
 * and Farach-Colton (2000) the four algorithms are
 *
 * 1. Direct Search  <O(1),       O(n)>
 * 2. Quadratic RMQ  <O(n^2),     O(1)>
 * 3. Sparse RMQ     <O(n log n), O(1)>
 * 4. Restricted RMQ <O(n),       O(1)>  [abs(a[i]-a[i-1]) == 1]
 * 5. RMQ            <O(n),       O(1)>  Solved in terms of LCA
 */

#include <stdio.h>
#include <stdlib.h>
#include "rmq.h"
#include "lca.h"
#include "util.h"

/**
 * Direct solution to the Range Minimum Query problem which does no
 * pre-processing, but takes time O(abs(j-i)) to answer a query be direct
 * search. 
 */
direct_rmq_t *direct_rmq_init( unsigned int *a, unsigned int n )
{
  direct_rmq_t *rmq = (direct_rmq_t *) malloc( sizeof( direct_rmq_t ));

  if ( rmq != NULL )
    rmq->a = a;

  return rmq;
}

unsigned int direct_rmq_exec( direct_rmq_t *rmq, unsigned int i, unsigned int j )
{
  unsigned int min_index;

  /* Make sure that i <= j */
  if ( j < i )
    uint_swap( &i, &j );

  /* Search for the index of the minimum element */
  for ( min_index = i; i <= j; i++ )
    if ( rmq->a[i] < rmq->a[min_index] )
      min_index = i;

  return min_index;
}

/****************************************************************************
 *
 * Naive solution to the Range Minimum Query problem which answers a query
 * in constant time, but uses quadratic space.  A trivial dynamic
 * programming approach is used to avoid cubic processing time.
 *
 * The solution matrix is symmetric, so we only allocate a diagonal
 * matrix to save half of our memory. 
 *
 ***************************************************************************/

quadratic_rmq_t *quadratic_rmq_init( unsigned int *a, unsigned int n )
{
  unsigned i, j, **M = malloc_upper_matrix( n );
  quadratic_rmq_t *rmq;

  /* The diagonal corresponds to RMQ[i:i] which is always element i. Also,
   * the RMQ array is symmetric and each RMQ answer is solved by the
   * following smaller problem. 
   *
   *  M[i:j] = argmin( a[M[i:j-1]], a[j] )
   */
  for ( i = 0; i < n; i++ )
  {
    M[i][i] = i;
    for ( j = i+1; j < n; j++ )
      M[i][j] = uint_argmin( a, M[i][j-1], j );
  }

  /**
   * Allocate the structure and return
   */
  rmq    = (quadratic_rmq_t *) malloc( sizeof( quadratic_rmq_t ));
  rmq->M = M;

  return rmq;
}

/**
 * Perform a RMQ by direct table lookup
 */
unsigned int quadratic_rmq_exec( quadratic_rmq_t *rmq, unsigned int i,
                                 unsigned int j )
{
  /* Make sure that i <= j */
  if ( j < i )
    uint_swap( &i, &j );

  return rmq->M[i][j];
}

/****************************************************************************
 *
 * A faster general RMQ solution that takes space O(n log n) and can
 * perform queries in O(1) time.  We precompute (via dynamic programming)
 * each query whose length is a power of two. e.g.
 *
 * for i in range 1:n
 *  for j in range 1:log n
 *   M[i,j] = minimum in a[i:i+2^j]
 *
 ***************************************************************************/

sparse_rmq_t *sparse_rmq_init( unsigned int *a, unsigned int n )
{
  unsigned int i, j, blk_size, base_two = log_base_2( n );
  unsigned int **M = malloc_unsigned_matrix( base_two + 1, n );
  sparse_rmq_t *rmq;

  /* Fill in the first row with the individual elements */
  for ( i = 0; i < n; i++ )
    M[0][i] = i;

  /**
   * Build up the rest of the array using dynamic programming.  We split
   * the inner loop into two pieces to avoid having to constantly do a
   * bounds check.
   */
  for ( blk_size = 2, i = 1; i <= base_two; i++, blk_size *= 2 )
  {
    for ( j = 0; j < n - blk_size + 1; j++ )
      M[i][j] = uint_argmin( a, M[i-1][j], M[i-1][j + blk_size/2] );

    for ( ; j < n; j++ )
      M[i][j] = M[i-1][j];
  }

  /**
   * Allocate the structure and return
   */
  rmq    = (sparse_rmq_t *) malloc( sizeof( sparse_rmq_t ));
  rmq->M = M;
  rmq->a = a;

  return rmq;
}

/**
 * Perform the query by finding the largest power of 2 that fits within the
 * query range, lookup up the minimum for each overlapping half of the
 * range and then choosing the smaller of the two.
 */
unsigned int sparse_rmq_exec( sparse_rmq_t *rmq, unsigned int i, 
                              unsigned int j )
{
  unsigned int k, span;

  /* Make sure that i <= j */
  if ( j < i )
    uint_swap( &i, &j );

  /* If the query is degenerate, return immediately ( log(0) undefined ) */
  if ( i == j )
    return i;

  k    = log_base_2( j - i );
  span = ( 1 << k );

  return uint_argmin( rmq->a, rmq->M[k][i], rmq->M[k][j - span + 1] );
}

/****************************************************************************
 *
 * Solves the +/- 1 Range Minimum Query problem in linear preprocessing
 * time and constant query time.
 *
 * This is the most complicated RMQ algorithm, even though it does run very
 * quickly (asymptotically).
 *
 ***************************************************************************/

pm_rmq_t *pm_rmq_init( unsigned int *a, unsigned int n )
{
  unsigned int *A, *B;
  unsigned int i, j, max_blocks, block_size, block_len;
  unsigned int min_idx, start, finish, block_id;
  sparse_rmq_t *block_range;
  quadratic_rmq_t **blocks, **pre;
  pm_rmq_t *rmq;

  /**
   * Compute the size of each parition block.  The last block may have a
   * different size and will have to be handled separately.
   */
  block_size = log_base_2( n ) / 2;
  block_len  = (n + block_size - 1) / block_size;
  max_blocks = 1 << ( block_len - 1 );

  /**
   * Allocate the arrays we will need
   */
  A   = (unsigned int *) malloc( sizeof( unsigned int ) * block_len );
  B   = (unsigned int *) malloc( sizeof( unsigned int ) * block_len );
  pre = (quadratic_rmq_t **) calloc( max_blocks, sizeof(quadratic_rmq_t *));
  blocks = (quadratic_rmq_t **) malloc( sizeof( quadratic_rmq_t * ) * block_len );

  /**
   * Find the minimum value and the position of that value in each block.
   * We also compute the normalized block number of which there are, at
   * most, O(sqrt(n)) different types.
   */
  for ( i = start = 0; i < block_len; i++, start += block_size )
  {    
    finish   = uint_min( start + block_size, n );
    min_idx  = start;
    block_id = 0;

    for ( j = start + 1; j < finish; j++ )
    {
      /* Find the minimum value */
      min_idx = uint_argmin( a, min_idx, j );

      /* Build up the block id -- a binary number where 0 == +1, 1 == -1 */
      block_id  = 2 * block_id;
      block_id += ( a[j] > a[j-1] ) ? 1 : 0; 
    }

    /* Set the minimum value and index for this block */
    A[i] = a[min_idx];
    B[i] = min_idx - start;

    /* Check to see if a solution has been precomputed for this block */
    if ( pre[block_id] == NULL )
      pre[block_id] = quadratic_rmq_init( a + start, finish - start );

    /* Assign the proper precomputed solution to this block */
    blocks[i] = pre[block_id];
  }

  /**
   * Use the sparse RMQ algorithm on A. Which will take time O(k log k)
   * where k = 2n/log n.  Thus the total processing time is O(n)
   */
  block_range = sparse_rmq_init( A, block_len );

  /**
   * Allocate the structure and return
   */
  rmq = (pm_rmq_t *) malloc( sizeof( pm_rmq_t ));

  rmq->A           = A;
  rmq->B           = B;
  rmq->blocks      = blocks;
  rmq->block_range = block_range;
  rmq->a           = a;
  rmq->block_size  = block_size;

  free( pre );

  return rmq;
}

unsigned int pm_rmq_exec( pm_rmq_t *rmq, unsigned int i, unsigned int j )
{
  unsigned int block_size = rmq->block_size;
  unsigned int start, finish, base, pos;
  unsigned int iblk, jblk, i_rmq, j_rmq, s_rmq;

  /* Make sure that i <= j */
  if ( j < i )
    uint_swap( &i, &j );

  iblk = i / block_size;
  jblk = j / block_size;

  /**
   * If both indices are within the same block, we can use the quadratic RMQ
   * data structure to answer the query in constant time.
   */
  if ( iblk == jblk )
  {
    /* Get the offset within the block and do the RMQ */
    base   = iblk * block_size;
    start  = i - base;
    finish = start + j - i;

    /* Perform the query and map the index back to the full range */
    pos    = quadratic_rmq_exec( rmq->blocks[iblk], start, finish );
    pos   += base;
  }

  /**
   * If the indices are in different blocks, we need to perform three
   * different queries (all in constant time) and take the min of all three.
   *
   * 1. Find the min in the block from i to the end of the block
   * 2. Find the min in the block from the beginning to j
   * 3. Use the sparse RMQ to get the minimum of all the in-between blocks
   */
  else
  {
    /* Compute the RMQ in the right portion of the i.th block */
    base   = iblk * block_size;
    start  = i - base;
    i_rmq  = quadratic_rmq_exec( rmq->blocks[iblk], start, block_size - 1 );
    i_rmq += base;

    /* Compute the RMQ in the left portion of the j.th block */
    base   = jblk * block_size;
    finish = j - base;
    j_rmq  = quadratic_rmq_exec( rmq->blocks[jblk], 0, finish );
    j_rmq += base;
    
    /* Take the smaller of the two */
    pos    = uint_argmin( rmq->a, i_rmq, j_rmq );

    /**
     * If i and j are not in adjacent blocks, find the RMQ of the
     * in-between blocks and compare it against the RMQ found above.
     */
    if ( jblk - iblk > 1 )
    {
      s_rmq  = sparse_rmq_exec( rmq->block_range, iblk+1, jblk-1 );
      s_rmq  = ( s_rmq * block_size ) + rmq->B[s_rmq];

      pos    = uint_argmin( rmq->a, s_rmq, pos );
    }
  }
  
  return pos;
}

/****************************************************************************
 *
 * Solves the general Range Minimum Query problem in linear preprocessing
 * time and constant query time by a reduction to the Least Common Ancestor
 * (LCA) problem.
 *
 ***************************************************************************/

rmq_t *rmq_init( unsigned int *a, unsigned int n )
{
  rmq_t  *rmq    = (rmq_t *) malloc( sizeof( rmq_t ));;
  node_t *c_tree = build_cartesian_tree( a, n );
  
  rmq->lca = lca_init( c_tree, n );
  rmq->a   = a;

  return rmq;
}

unsigned int rmq_exec( rmq_t *rmq, unsigned int i, unsigned int j )
{
  return rmq->a[lca_query( rmq->lca, i, j )->value];
}
