#ifndef RMQ_H
#define RMQ_H

#include "lca.h"

/**
 * Structure to implement a direct RMQ search solution
 */
typedef struct direct_rmq_t
{
  unsigned int *a;
} direct_rmq_t;

/**
 * Structure to hold the pre-processed data for the naive solution to the
 * RMQ problem.  This takes up quadratic space.
 */
typedef struct quadratic_rmq_t
{
  unsigned int **M;          /* Triangular matrix holding the solution */ 
} quadratic_rmq_t;

/**
 * Structure to hold the pre-processed data for the sparse solution to the
 * RMQ problem.  This takes up O(n log n) space. We have to keep a
 * reference to the orginal data in order to choose the half contains the
 * correct answer as query time.
 */
typedef struct sparse_rmq_t
{
  unsigned int **M;          /* Array holding to solutions to the RMQ   */
  unsigned int *a;           /* A reference to the original data set    */
} sparse_rmq_t;

/**
 * Structure to hold the pre-processed data for the +/-1RMQ problem. This
 * is a restricted version of the RMQ problem which is leveraged to proved
 * a <O(n), O(1)> solution to the Least Common Ancestor problem.
 */
typedef struct pm_rmq_t
{
  unsigned int *A;           /* Hold the min value for each block       */
  unsigned int *B;           /* Hold the position of the min value in A */
  quadratic_rmq_t **blocks;  /* Quadratic solutions for each block      */
  sparse_rmq_t *block_range; /* Sparse RMQ solution for A               */
  unsigned int *a;           /* A reference to the original data set    */
  unsigned int block_size;   /* length of each sub-block */
} pm_rmq_t;

/**
 * Structure to hold the pre-processed data for the general RMQ problem.
 */
typedef struct rmq_t
{
  struct lca_t *lca;
  unsigned int *a;
} rmq_t;


/* Functions for a direct RMQ algorithm <O(1), O(n)> */
direct_rmq_t *direct_rmq_init( unsigned int *a, unsigned int n );
unsigned int  direct_rmq_exec( direct_rmq_t *rmq, unsigned int i, unsigned int j );
     
/* Functions for a naive RMQ algorithm <O(n^2), O(1)> */
quadratic_rmq_t *quadratic_rmq_init( unsigned int *a, unsigned int n );
unsigned int     quadratic_rmq_exec( quadratic_rmq_t *rmq, unsigned int i, unsigned int j );

/* Functions for a sparse RMQ algorithm <O(n log n), O(1)> */
sparse_rmq_t *sparse_rmq_init( unsigned int *a, unsigned int n );
unsigned int  sparse_rmq_exec( sparse_rmq_t *rmq, unsigned int i, unsigned int j );

/* Functions for the +/-1 RMQ algorithm <O(n), O(1)> */ 
pm_rmq_t    *pm_rmq_init( unsigned int *a, unsigned int n );
unsigned int pm_rmq_exec( pm_rmq_t *rmq, unsigned int i, unsigned int j );

/* Functions for the general RMQ algorithm <O(n), O(1)> */ 
rmq_t       *rmq_init( unsigned int *a, unsigned int n );
unsigned int rmq_exec( rmq_t *rmq, unsigned int i, unsigned int j );

#endif
